#include <minmax.h>
#include <LEDA/ugraph.h>
#include <Tim/graph.h>
#include "lbfs.h"
#include "interval.h"

char interval_version[] = "$Id: interval.cc,v 1.7 1996/08/22 21:24:36 tbm Exp $";

/*****************************************************************************/

vertex_order sweep[MAX_SWEEPS];

/*****************************************************************************/

//@Man:
bool umbrella (const labelled_graph &G, const vertex_order &order) {

/**

	For each vertex, we record the distance in the order to the last
	subsequent neighbour (0 if none), and the number of subsequent
	neighbours.
	The ordering is then an umbrella ordering if the two numbers are
	equal for every vertex.
	Runs in time $O(m+n)$.

*/

	node_array<int> max_distance (G, 0);
	node_array<int> num_subsequent (G, 0);

	edge e;
	node u, v;

	forall_edges (e, G) {
		// Ensure u < v
		if (order.position[G.source (e)] < order.position[G.target (e)]) {
			u = G.source (e);
			v = G.target (e);
		} else {
			v = G.source (e);
			u = G.target (e);
		}
		num_subsequent[u]++;
		max_distance[u] = max (max_distance[u], order.position[v] - order.position[u]);
	}

	forall_nodes (v, G) {
		if (num_subsequent[v] != max_distance[v]) {
			if (options.full_info) {
				u = order.nodes.inf (order.nodes.item (order.position[v] + max_distance[v] - 1));
				cout << "\n\t\tEdge {" << G[v].label << ","
					<< G[u].label
					<< "} flies over something\n";
			}
			return false;
		}
	}

	return true;
}

/*****************************************************************************/

//@Man:
bool flies (
	const labelled_graph &G,
	const lexicographic_order &order,
	const node v
) {

/**

	Determine if a vertex flies.
	That is, determine if it is an unvisited vertex is a candidate
	for the next vertex to visit and is adjacent to an unvisited
	vertex outside the candidate set.
	Runs in time $O(|N[v]|)$.

*/

	if (order.visited[v] || (order.eq_classes[v] != order.order.first ())) {
		// Vertex has been visited or is not in the candidate set
		return false;
	}

	node u;
	forall_adj_nodes (u, v) {
		if ((!order.visited[u]) && (order.eq_classes[u] != order.eq_classes[v])) {
			// Vertex u is an unvisited neighbour outside the
			// candidate set
			return true;
		}
	}

	return false;
}

/*****************************************************************************/

//@Man:
flies_status get_flies_status (
	const labelled_graph &G,
	const lexicographic_order &order,
	const node v
) {

/**

	Determine flies status (FLIES_FLIES, FLIES_NEIGHBOUR or FLIES_OK)
	of a vertex.
	Runs in time $O(|N[v]|)$.

*/

	if (flies (G, order, v)) {
		return FLIES_FLIES;
	}

	node u;
	forall_adj_nodes (u, v) {
		if (flies (G, order, u)) {
			return FLIES_NEIGHBOUR;
		}
	}

	return FLIES_OK;
}

/*****************************************************************************/

//@Man:
node find_last_candidate (
	const graph &G,
	const lexicographic_order &lex_order,
	vertex_order &order
) {

/**

	Find the last candidate vertex in an order.
	Currently has an amortized run time of $O(n^2)$.
	Will hopefully have an amortized run time of $O(m+n)$.

*/

	if (lex_order.order.empty ()) {
		return NULL;
	}

	// Find candidate set
	node v = lex_order.order.head ()->S.head ();
	list_item candidates = lex_order.eq_classes[v];

	// Innefficient as yet
	list_item item;
	for (item = order.nodes.last (); item != NULL; item = order.nodes.pred (item)) {
		v = order.nodes.contents (item);
		if ((!lex_order.visited[v]) && (lex_order.eq_classes[v] == candidates)) {
			return v;
		}
	}
	return NULL; // This will never happen
}

/*****************************************************************************/

//@Man:
node picklast (
	const labelled_graph &G,
	const lexicographic_order &order,
	const list<node> &choices,
	void *info1,
	void *info2
) {

/**

	Tiebreaker that chooses in favour of later vertices in some order.
	Currently has an amortized run time of $O(n^2)$.
	Will hopefully have an amortized run time of $O(m+n)$.

*/

	if (choices.size () <= 1) {
		return choices.head ();
	}

	return find_last_candidate (G, order, *((vertex_order*)info1));
}

/*****************************************************************************/

flies_scheme flies_strong = {
	{ PICK_BETA,  PICK_BETA,  PICK_BETA },
	{ PICK_ALPHA, PICK_BETA,  PICK_BETA },
	{ PICK_ALPHA, PICK_BETA,  PICK_BETA }
};

flies_scheme flies_weak = {
	{ PICK_BETA,  PICK_BETA,  PICK_BETA },
	{ PICK_BETA,  PICK_BETA,  PICK_BETA },
	{ PICK_ALPHA, PICK_ALPHA, PICK_BETA }
};

flies_scheme flies_mixed = {
	{ PICK_BETA,  PICK_BETA,  PICK_BETA },
	{ PICK_ALPHA, PICK_BETA,  PICK_BETA },
	{ PICK_ALPHA, PICK_ALPHA, PICK_BETA }
};

/*****************************************************************************/

//@Man:
node picknonflier (
	const labelled_graph &G,
	const lexicographic_order &order,
	const list<node> &choices,
	void *info1,
	void *info2
) {

/**

	Tiebreaker that chooses in favour of later non-flier in two orders.
	Currently has an amortized run time of $O(n^2)$.
	Will hopefully have an amortized run time of $O(m+n)$.

*/

	if (choices.size () <= 1) {
		return choices.head ();
	}

	node beta = find_last_candidate (G, order, *((vertex_order*)info1));
	node alpha = find_last_candidate (G, order, *((vertex_order*)info2));

	flies_status beta_status  = get_flies_status (G, order, beta);
	flies_status alpha_status = get_flies_status (G, order, alpha);

	if (options.full_info) {
		cout << "\t\t\tbeta=" << G[beta].label << " " << beta_status << "\n";
		cout << "\t\t\talpha=" << G[alpha].label << " " << alpha_status << "\n";
	} else {
		// If full_info is set, we're re-running a case, so
		// don't count the tiebreaker hits twice
		statistics.count (G, alpha, alpha_status, beta, beta_status);
	}

	if (options.flies [alpha_status][beta_status] == PICK_ALPHA) {
		return alpha;
	} else {
		return beta;
	}
}

/*****************************************************************************/

//@Man:
bool old (
	labelled_graph &G,
	node v,
	tiebreaker first_tb,
	void *tbinfo1,
	void *tbinfo2
) {

/**

	Old four-sweep algorithm.  This is known to fail, but is good for
	testing purposes.
	Currently runs in time $O(n^2)$ (because of tiebreakers).
	Will hopefully run in time $O(n+m)$.

*/

	sweep[0].init (G);
	sweep[1].init (G);
	sweep[2].init (G);
	sweep[3].init (G);
	if (v == NULL) {
		v = G.choose_node ();
	}
	v = LBFS (G, v, sweep[0], first_tb, tbinfo1, tbinfo2);
	v = LBFS (G, v, sweep[1], picklast, &sweep[0]);
	v = LBFS (G, v, sweep[2], picklast, &sweep[1]);
	LBFS (G, v, sweep[3], picklast, &sweep[2]);
	return umbrella (G, sweep[3]);
}

/*****************************************************************************/

//@Man:
bool three_sweep (
	labelled_graph &G,
	node v,
	tiebreaker first_tb,
	void *tbinfo1,
	void *tbinfo2
) {

/**

	New algorithm using the tiebreaker based on flying vertices.
	Currently runs in time $O(n^2)$ (because of tiebreakers).
	Will hopefully run in time $O(n+m)$.

*/
	sweep[0].init (G);
	sweep[1].init (G);
	sweep[2].init (G);
	if (v == NULL) {
		v = G.choose_node ();
	}
	v = LBFS (G, v, sweep[0], first_tb, tbinfo1, tbinfo2);
	v = LBFS (G, v, sweep[1], picklast, &sweep[0]);
	LBFS (G, v, sweep[2], picknonflier, &sweep[1], &sweep[0]);
	return umbrella (G, sweep[2]);
}

/*****************************************************************************/

//@Man:
bool five_sweep (
	labelled_graph &G,
	node v,
	tiebreaker first_tb,
	void *tbinfo1,
	void *tbinfo2
) {

/**

	New algorithm using the tiebreaker based on flying vertices.
	Currently runs in time $O(n^2)$ (because of tiebreakers).
	Will hopefully run in time $O(n+m)$.
	Checks if pass 3 produces the same order as pass 5.

*/
	sweep[0].init (G);
	sweep[1].init (G);
	sweep[2].init (G);
	sweep[3].init (G);
	sweep[4].init (G);
	if (v == NULL) {
		v = G.choose_node ();
	}
	v = LBFS (G, v, sweep[0], first_tb, tbinfo1, tbinfo2);
	v = LBFS (G, v, sweep[1], picklast, &sweep[0]);
	v = LBFS (G, v, sweep[2], picknonflier, &sweep[1], &sweep[0]);
	v = LBFS (G, v, sweep[3], picknonflier, &sweep[2], &sweep[1]);
	LBFS (G, v, sweep[4], picknonflier, &sweep[3], &sweep[2]);
	return umbrella (G, sweep[4]);
}

/*****************************************************************************/

//@Man:
bool four_plus (
	labelled_graph &G,
	node v,
	tiebreaker first_tb,
	void *tbinfo1,
	void *tbinfo2
) {

/**

	New algorithm using the tiebreaker based on flying vertices.
	Currently runs in time $O(n^2)$ (because of tiebreakers).
	Will hopefully run in time $O(n+m)$.

*/
	sweep[0].init (G);
	sweep[1].init (G);
	sweep[2].init (G);
	sweep[3].init (G);
	if (v == NULL) {
		v = G.choose_node ();
	}
	v = LBFS (G, v, sweep[0], first_tb, tbinfo1, tbinfo2);
	v = LBFS (G, v, sweep[1], picklast, &sweep[0]);
	v = LBFS (G, v, sweep[2], picklast, &sweep[1]);
	LBFS (G, v, sweep[3], picknonflier, &sweep[2], &sweep[1]);
	return umbrella (G, sweep[3]);
}

/*****************************************************************************/

//@Man:
bool four_star (
	labelled_graph &G,
	node v,
	tiebreaker first_tb,
	void *tbinfo1,
	void *tbinfo2
) {

/**

	New algorithm using the tiebreaker based on flying vertices.
	Currently runs in time $O(n^2)$ (because of tiebreakers).
	Will hopefully run in time $O(n+m)$.

*/
	sweep[0].init (G);
	sweep[1].init (G);
	sweep[2].init (G);
	sweep[3].init (G);
	if (v == NULL) {
		v = G.choose_node ();
	}
	v = LBFS (G, v, sweep[0], first_tb, tbinfo1, tbinfo2);
	v = LBFS (G, v, sweep[1], picklast, &sweep[0]);
	v = LBFS (G, v, sweep[2], picknonflier, &sweep[1], &sweep[0]);
	LBFS (G, v, sweep[3], picknonflier, &sweep[2], &sweep[1]);
	return umbrella (G, sweep[3]);
}
