#include <LEDA/node_array.h>

/*****************************************************************************/

//@Man:
class vertex_order {
private:
	/// The graph
	const labelled_graph *the_graph;

public:
	/// List of nodes
	list<node> nodes;
	/// Node positions
	node_array<int> position;
	/// List entries
	node_array<list_item> position_item;

	///				\hfill $O(n)$
	inline void init (const labelled_graph &G) { position.init (G); position_item.init (G); nodes.clear (); the_graph = &G; };
	///				\hfill $O(n)$
	inline vertex_order (const labelled_graph &G) { init (G); };
	///				\hfill $O(n)$
	inline vertex_order () { the_graph = NULL; };
	///				\hfill $O(n)$
	friend ostream & operator << (ostream&, const vertex_order&);
	///
	friend bool operator == (const vertex_order&, const vertex_order&); 
};

/*****************************************************************************/

class lexicographic_order;

//@Man:
//@{
typedef node tiebreaker (
	/// The graph being traversed
	const labelled_graph &G,
	/// The ordering so far
	const lexicographic_order &order, 
	/// Candidates for next vertex to visit
	const list<node> &choices,
	/// Extra info
	void *info1 = NULL,
	/// Extra info
	void *info2 = NULL
);

/// Tiebreaker that chooses arbitrarily from the choices
tiebreaker arbitrary;
//@}

/*****************************************************************************/

//@Man:
class eq_class {
private:
	/// The graph
	const labelled_graph *the_graph;

public:
	/// The refinement step when this set was created
        int creation_timestamp;
	/// The nodes in the class
	list<node> S;

	///				\hfill	$O(1)$
	inline eq_class (const labelled_graph &G, int timestamp = 0) {
		the_graph = &G;
		creation_timestamp = timestamp;
	};
	///				\hfill $O(n)$
	friend ostream & operator << (ostream&, const eq_class&);
	
	LEDA_MEMORY(eq_class)
};

/**

	The lexicographic ordering of the nodes is stored as a linked list
	of equivalence classes.  Initially, no information about the order
	is known, so all of the nodes are in a single set.  When a node v
	is visited, we refine each class by partitioning it into the set
	of vertices adjacent to v and the set of vertices not adjacent to v.

*/

/*****************************************************************************/

//@Man:
class lexicographic_order {
private:
	/// The graph
	const labelled_graph *the_graph;

public:
	/// Number of refinements already performed
        int refinements;
	/// List of equivalence classes
	list<eq_class*>	order;
	/// Has the vertex been visited?
	node_array<bool> visited;
	/// Which eq class a vertex is in
	node_array<list_item> eq_classes;
	/// Pointer to entry in eq class
	node_array<list_item> eq_class_entry;

	/// Initialization		\hfill $O(n)$
	void init (const labelled_graph&);
	///				\hfill $O(n)$
        inline lexicographic_order (const labelled_graph &G) { init (G); };

	/// Refine ordering		\hfill $O(|list|)$
        void refine (const list<node>);
	/// Extract node		\hfill $O(1)$
        void extract (node);
	/// Extract next node		\hfill $O(1+tiebreaker)$
        node extract (tiebreaker = arbitrary, void *tbinfo1 = NULL, void *tbinfo2 = NULL);

	///				\hfill $O(n)$
	friend ostream & operator << (ostream&, const lexicographic_order&);
};

/**

	Ideally, much of the data should be private, but this makes life
	simpler for some the the AT-free routines, such as the fly checking.

*/

/*****************************************************************************/

//@Man:
node LBFS (
	/// Graph being searched
	const labelled_graph &G,
	/// Starting node of search
	node v,
	/// Ordering produced
	vertex_order &order,
	/// Tiebreaker
	tiebreaker tb = arbitrary,
	/// Optional tiebreaker parameter
	void *tbinfo1 = NULL,
	/// Optional tiebreaker parameter
	void *tbinfo2 = NULL
);

/**

	Performs a lexicographic breadth-first search of $G$, starting
	at vertex $v$ and using $tb$ to break ties.
	Returns the last node visited.
	The running time is $O(m+n+T(m,n))$ where $T(m,n)$ is the total
	run time of the $n$ calls to the tiebreaker.

*/
